;*********************************************************************
; FAT CONTROLLER ~ MIDI to CV Converter                              *
;*********************************************************************
; (c) 2002 SM-Sound / Seb Francis                                    *
;*********************************************************************
;
; Subroutines for controlling LCD display
;
;*********************************************************************


	
; --------------------------------
; LCD Functions
; Located in bank 0
; --------------------------------
	
	
; Waits while LCD is busy
; Must be called with RAM bank 0 selected
LCD_WaitWhileBusy macro
	bcf	PORTB,LCDRS		; Setup LCD for reading IR
	bsf	PORTB,LCDRW
	
	bsf	STATUS,RP0		; Select RAM bank 1
	movlw	b'11111111'		; switch port D to inputs
	movwf	TRISD
	bcf	STATUS,RP0		; Select RAM bank 0
	
	bsf	PORTB,LCDE		; LCD Enable
	nop
	movfw	PORTD			; Read byte into W
	movwf	LCD_VALUE_RD		; Write byte into LCD_VALUE_RD
	
	bcf	PORTB,LCDE		; LCD Disable
	
	btfsc	LCD_VALUE_RD,BUSY	; Wait while busy flag set
	goto	$-6
	
	bsf	STATUS,RP0		; Select RAM bank 1
	clrf	TRISD			; Switch port D back to outputs
	bcf	STATUS,RP0		; Select RAM bank 0
endm
	
	
; --------------------------------



; Writes contents of W to the LCD IR, doesn't check busy flag
; Must be called with RAM bank 0 selected
LCD_WriteIR_NoWait
	bcf	PORTB,LCDRS
	bcf	PORTB,LCDRW
	bsf	PORTB,LCDE
	movwf	PORTD
	nop
	bcf	PORTB,LCDE
	return
	
	
; --------------------------------


; Writes contents of W to the LCD IR
; Must be called with RAM bank 0 selected
LCD_WriteIR
	movwf	LCDTEMP1
	LCD_WaitWhileBusy
	bcf	PORTB,LCDRS
	bcf	PORTB,LCDRW
	bsf	PORTB,LCDE
	movfw	LCDTEMP1
	movwf	PORTD
	nop
	bcf	PORTB,LCDE
	return

	
; --------------------------------


; Writes contents of W to the LCD DR
; Must be called with RAM bank 0 selected
LCD_WriteDR
	movwf	LCDTEMP1
	LCD_WaitWhileBusy
	bsf	PORTB,LCDRS
	bcf	PORTB,LCDRW
	bsf	PORTB,LCDE
	movfw	LCDTEMP1
	movwf	PORTD
	nop
	bcf	PORTB,LCDE
	return
	
	
; --------------------------------
	
	
; Initialises LCD
; Must be called with RAM bank 0 selected
LCD_Init
	movlw	b'00110000'		; Init instruction
	call	LCD_WriteIR_NoWait
	
IFDEF	ENABLE_DELAYS
	; Wait more than 4.1ms
	wait_ms	D'5'
ENDIF
	
	movlw	b'00110000'		; Init instruction
	call	LCD_WriteIR_NoWait

IFDEF	ENABLE_DELAYS
	; Wait more than 100us
	wait_us D'120'
ENDIF
	
	movlw	b'00110000'		; Init instruction
	call	LCD_WriteIR_NoWait
	
	movlw	b'00111000' 		; Set 8-bit, 2 lines, small font
	call	LCD_WriteIR
	
	movlw	b'00001100' 		; Display on, cursor off, blinking off
	call	LCD_WriteIR
	
	movlw	b'00000001' 		; Clear display
	call	LCD_WriteIR
	
	movlw	b'00000110' 		; Increment after each letter, do not shift
	call	LCD_WriteIR
	
	return
	
	
; --------------------------------


; Writes 8 character string table pointed to by W to top row of LCD
; W should contain the string address / 8
; Must be called with RAM bank 0 selected
;
; Example:
; --------
; movlw	sBASE_CHA / 8
; call	LCD_WriteTopRowString

LCD_WriteTopRowString
	movwf   LCDTEMP2		; Save string pointer
	
	movlw	0x80 | 0x00		; Set character position to start of top row
	call	LCD_WriteIR
	
	goto	LCD_WriteStringFromTEMP2
	

; --------------------------------


; Writes 8 character string table pointed to by W to bottom row of LCD
; W should contain the string address / 8
; Must be called with RAM bank 0 selected
;
; Example:
; --------
; movlw	sOff / 8
; call	LCD_WriteBottomRowString

LCD_WriteBottomRowString
	movwf   LCDTEMP2		; Save string pointer
	
	movlw	0x80 | 0x40		; Set character position to start of bottom row
	call	LCD_WriteIR
	
	goto	LCD_WriteStringFromTEMP2
	

; --------------------------------


LCD_StartBottomRow
	movlw	0x80 | 0x40		; Set character position to start of bottom row
	goto	LCD_WriteIR
	

; --------------------------------


LCD_BinaryToBcd macro
	clrf	hundreds
	swapf	bin,w
	addwf	bin,w
	andlw	B'00001111'
	btfsc	STATUS,DC
	 addlw	0x16
	btfsc	STATUS,DC
	 addlw	0x06
	addlw	0x06
	btfss	STATUS,DC
	addlw	-0x06
	btfsc	bin,4
	 addlw	0x16 - 1 + 0x6
	btfss	STATUS,DC
	 addlw	-0x06
	btfsc	bin,5
	 addlw	0x30
	btfsc	bin,6
	 addlw	0x60
	btfsc	bin,7
	 addlw	0x20
	addlw	0x60
	rlf	hundreds,f
	btfss	hundreds,0
	 addlw	-0x60
	movwf	tens_and_ones
	btfsc	bin,7
	 incf	hundreds,f
endm


; Writes 8 bit integer in W in decimal to LCD
; Must be called with RAM bank 0 selected
;
; Example:
; --------
; movlw	D'15'
; call	LCD_WriteInt

bin		EQU	LCDTEMP1
hundreds	EQU	LCDTEMP2
tens_and_ones	EQU	LCDTEMP3

LCD_WriteInt
	movwf	bin
	LCD_BinaryToBcd			; Convert integer to 3 nybble BCD
	
	movfw	hundreds		; Step through digits until non-zero digit
	jnz	LCD_WriteInt_Hundreds
	swapf	tens_and_ones,w
	andlw	0xF
	jnz	LCD_WriteInt_Tens
	goto	LCD_WriteInt_Ones
	
LCD_WriteInt_Hundreds
	addlw	"0"			; Convert digit to ASCII
	call	LCD_WriteDR		; Write character to LCD
	
	swapf	tens_and_ones,w
	andlw	0xF
LCD_WriteInt_Tens
	addlw	"0"			; Convert digit to ASCII
	call	LCD_WriteDR		; Write character to LCD

LCD_WriteInt_Ones
	movf	tens_and_ones,w
	andlw	0xF
	addlw	"0"			; Convert digit to ASCII
	goto	LCD_WriteDR		; Write character to LCD and return
	

; --------------------------------


; Writes 16 bit integer at FSR and FSR+1 in decimal to LCD
; Must be called with RAM bank 0 selected

TENKS		EQU	LCDTEMP1
THOUS		EQU	LCDTEMP2
HUNDS		EQU	LCDTEMP3
TENS		EQU	LCDTEMP4
UNITS		EQU	LCDTEMP5

LCD_WriteLongInt
	call	LCD_LongBinaryToBcd	; Convert integer to 5 nybble BCD
	
	movfw	TENKS			; Step through digits until non-zero digit
	jnz	LCD_WriteLongInt_TENKS
	movfw	THOUS
	jnz	LCD_WriteLongInt_THOUS
	movfw	HUNDS
	jnz	LCD_WriteLongInt_HUNDS
	movfw	TENS
	jnz	LCD_WriteLongInt_TENS
	movfw	UNITS
	jnz	LCD_WriteLongInt_UNITS
	return				; All 0s, just return
	
LCD_WriteLongInt_TENKS
	addlw	"0"			; Convert digit to ASCII
	call	LCD_WriteDR		; Write character to LCD
	
	movfw	THOUS
LCD_WriteLongInt_THOUS
	addlw	"0"			; Convert digit to ASCII
	call	LCD_WriteDR		; Write character to LCD
	
	movfw	HUNDS
LCD_WriteLongInt_HUNDS
	addlw	"0"			; Convert digit to ASCII
	call	LCD_WriteDR		; Write character to LCD
	
	movfw	TENS
LCD_WriteLongInt_TENS
	addlw	"0"			; Convert digit to ASCII
	call	LCD_WriteDR		; Write character to LCD

	movfw	UNITS
LCD_WriteLongInt_UNITS
	addlw	"0"			; Convert digit to ASCII
	goto	LCD_WriteDR		; Write character to LCD and return


; --------------------------------


; Writes 16 bit integer at FSR and FSR+1, divided by 100, in decimal to LCD
; Must be called with RAM bank 0 selected

LCD_WriteLongIntDiv100
	call	LCD_LongBinaryToBcd	; Convert integer to 5 nybble BCD
	
	movfw	TENKS			; Step through digits until non-zero digit
	jnz	LCD_WriteLongIntDiv100_TENKS
	movfw	THOUS
	jnz	LCD_WriteLongIntDiv100_THOUS
	goto	LCD_WriteLongIntDiv100_HUNDS
	
LCD_WriteLongIntDiv100_TENKS
	addlw	"0"			; Convert digit to ASCII
	call	LCD_WriteDR		; Write character to LCD
	
	movfw	THOUS
LCD_WriteLongIntDiv100_THOUS
	addlw	"0"			; Convert digit to ASCII
	call	LCD_WriteDR		; Write character to LCD
	
LCD_WriteLongIntDiv100_HUNDS
	movfw	HUNDS
	addlw	"0"			; Convert digit to ASCII
	call	LCD_WriteDR		; Write character to LCD
	
	movlw	"."			; Decimal place
	call	LCD_WriteDR
	
	movfw	TENS
	addlw	"0"			; Convert digit to ASCII
	call	LCD_WriteDR		; Write character to LCD

	movfw	UNITS
	addlw	"0"			; Convert digit to ASCII
	goto	LCD_WriteDR		; Write character to LCD and return

	
; ***********************************************************************************************
; DESCRIPTION: To evaluate the number of TENKS (10K or 10,000), THOUS, HUNDS, TENS and UNITS 
; in a 16-bit natural binary number --- Maximum value FFFFh -> 6 TENKS (10^4), 5 THOUS (10^3)
; 5 HUNDS (10^2), 3 TENS (10^1) and 5 UNITS (10^0).
; A 16-bit binary number can be written as a 4-digit hexadecimal number:
; NUMBER = A3*16^3 + A2*16^2 + A1*16^1 + A0*16^0            .....(1) or
; NUMBER = A3*4096 + A2*256  + A1*16   + A0                 .....(2) or in BCD as:
; NUMBER = B4*10^4 + B3*10^3 + B2*10^2 + B1*10^1 + B0*10^0  .....(3) or
; NUMBER = B4*10000+ B3*1000 + B2*100  + B1*10   + B0       .....(4)
; In a nutshell, to convert from binary/hex to decimal we must evlauate the B[4..0] coefficients
;
; One way of doing this is to collect together all thousands, hundreds, tens and units
; THOUS = B3 = 4*A3                         (from the 4000 part of the 16^4 = 4096 weighting)
; HUNDS = B2 = 2*A2                         (from the 200 part of the 16^3 = 256 weighting)
; TENS  = B1 = (9*A3)+(5*A2)+(1*A1)         (from the 9 part of 4096, 5 part of 256 & 1 part of 16)
; UNITS = B0 = 6*(A3+A2+A1)+A0              (from units part of weightings)
; Now we have:
; THOUS = 4*A3                   : Range 0 -> 60 (4*15)
; HUNDS = 2*A2                   : Range 0 -> 30
; TENS  = 9*A3 + 5*A2 + A1       : Range 0 -> 225
; UNITS = 6*(A3+A2+A1) + A0      : Range 0 -> 285
; where the maximum values are derived remembering that a hex digit's maximum value is 15 (Fh).
;
; Now each BCD digit can only go up to 9, so although the values for each decade above is valid,
; each BCD digit needs to be normalized to lie in the range 0 -> 9.  This could be done by dividing each 
; digit from UNITS upwards in turn by ten.  The remainder from each division (the modulus) is 0 to 9 
; as required.  In order to keep the overall value of the expression the same the quotient
; (in the case of UNITS, this is the number of tens) can be added to the next highest decade
; to leave the overall total the same (that is the number of tens removed from the units decade
; is added to the tens decase to balance the process).  Once the UNITS value has been normalized
; the process is continued for the TENS count the the quotient being subtracted from HUNDS etc.
;
; Rather than divide by ten, if we can arrange for the B[3..0] coefficients of equation 2 to be
; always negative then we can simply normalise each decade n by continually adding ten until
; the coefficient turns positive (signalled by a Carry) while at the same time decrementing the n+1 decade
; to balance.  This "normalize and compensate" process is continued until all decades have digits 
; from 0 thru 9.
; As a first step towards the goal we can rewrite equation 2 as: (why not?)
; NUMBER = A3*(4100-4) + A2*(260-4) + A1*(20-4) + A0          .....5 
; Now grouping as before gives
; THOUS = 4*A3                     : Range 0   -> 60
; HUNDS = A3 + 2*A2                : Range 0   -> 45
; TENS  = (6*A2) + (2*A1)          : Range 0   -> 120
; UNITS = A0 - 4*(A3+A2+A1)        : Range -60 -> 15
; This is not much of an improvement but we can further tweak the coefficients provided that the 
; overall value is not changed.  Thus:
; For B0 (UNITS) we subtract twenty as the first higher decade above 15, and we -2 from B1 to balance
; For B1 (TENS) we subtract 140 and add two giving -138
; For B2 (HUNDS) we subtract 60 and add 14 (to compensate for the -140 from B1 (TENS)
; For B3 (THOUS) we subtract 70 and add 6 to compensate the -60 from B2 (HUNDS)
; For B4 (TENKS) we add 7 to compensate the -70 from B3 (THOUS)
;
; Our final list of coefficients now is:
; UNITS = A0 - 4*(A3+A2+A1) - 20   : Range -80  -> -5
; TENS  = (6*A2) + (2*A1) -138     : Range -138 -> -18
; HUNDS = A3 + 2*A2 - 46           : Range -46  -> -1
; THOUS = 4*A3 - 64                : Range -64  -> -4
; TENKS = 7                        ; Does not have to be normallized!

; The first part of the subroutine below determines these initial values of the five coefficients
; The second part normalizes each decade
; * ENTRY  : High byte at FSR+1, low byte at FSR                                                 *
; * EXIT   : Files TENKS, THOUS, HUNDS, TENS, UNITS have their appropriate values 0 -> 9         *
; * EXAMPLE: NUMH = FFh, NUML = FFh: TENKS = 06, THOUS = 05, HUNDS = 05, TENS = 03, UNITS = 05   *
; ************************************************************************************************

LCD_LongBinaryToBcd
	incf	FSR
	
	swapf	INDF,w		; W = A2:A3
	iorlw	b'11110000'	; 1111:A3 is A3 - 16 (F0h is 2's complement of 16)
	movwf	THOUS		; THOUS = A3 - 16
	addwf	THOUS,f		; Adding the same again gives THOUS = 2*A3 - 32
	addlw	-d'30'		; (A3 - 16) - 30
	movwf	HUNDS		; HUNDS = (A3 - 46)
	addlw	d'50'		; (A3 - 46) + 50
	movwf	UNITS		; UNITS = A3 + 4

	movf	INDF,w		; Get upper byte
	andlw	b'00001111'	; Isolate the lower nybble A2
	addwf	HUNDS,f		; HUNDS = (A3 - 46) + A2
	addwf	HUNDS,f		; HUNDS = (A3 - 46) + A2 + A2 = 2*A2 + (A3 - 46)
	addwf	UNITS,f          ; UNITS = (A3 + 4) + A2
	addlw	-d'23'		; A2 - 23
	movwf	TENS		; TENS = A2 - 23
	addwf	TENS,f		; TENS = (A2 - 23) + (A2 - 23) 
	addwf	TENS,f		; TENS = (2*A2 -46) + (A2 - 23) = 3*A2 - 69
	
	decf	FSR

	swapf	INDF,w
	andlw	b'00001111'	; A1
	addwf	TENS,f		; TENS = A1 + (3*A2 - 69)
	addwf	UNITS,f		; UNITS = A1 + (A3 + A2 + 4) (no Carry)
	rlf	TENS,f		; multiply by 2 gives TENS = 2*A1 + 6*A2 - 138 (generates a Carry)  
	rlf	UNITS,f		; x2 + 1 (Carry) gives UNITS =  2*(A3 + A2 + A1) + 9
	comf	UNITS,f		; 1's complement of K gives -(K+1) gives -2*(A3 + A2 + A1) - 10
	rlf	UNITS,f		; x2 gives UNITS = -4*(A3 + A2 + A1) - 20

	movf	INDF,w
	andlw	b'00001111'	; A0
	addwf	UNITS,f		; UNITS = A0 -4*(A3 + A2 + A1) -20
	rlf	THOUS,f		; THOUS = 2*A3 - 32

	movlw	b'00000111'
	movwf	TENKS		; TENKS = 07

; Now we have all the coefficients in place, normalize each decade starting at UNITS
; by repetitively adding ten until the negative coefficient becomes positive
; This will occur when the 1s at the top of the binary byte (it is negative) 
; become 0s and this will generate a Carry out (eg. 11111000 (-8) + 00001010 (+10)
; -> 1,00000010 (+2 Carry 1)

	movlw	d'10'		
LB1	addwf	UNITS,f		; Keep adding ten units
	decf	TENS,f		; and decrementing the next decade to keep balance
	btfss	STATUS,C	; until there is a Carry out (overflow)
	 goto	LB1

LB2	addwf	TENS,f		; Repeat for the tens decade
	decf	HUNDS,f
	btfss	STATUS,C
	 goto	LB2

LB3	addwf	HUNDS,f		; Repeat for the hundreds decade
	decf	THOUS,f
	btfss	STATUS,C
	 goto	LB3

LB4	addwf	THOUS,f		; and repeat for the thousands decade
	decf	TENKS,f
	btfss	STATUS,C
	 goto	LB4

	return       


; --------------------------------


; Writes W spaces to LCD
; Must be called with RAM bank 0 selected
;
; Example:
; --------
; movlw	5
; call	LCD_WriteSpaces

LCD_WriteSpaces
	movwf	LCDTEMP2
	movlw	" "
	call	LCD_WriteDR
	decfsz	LCDTEMP2
	goto	$-2
	return


; --------------------------------
	

LCD_WriteStringFromTEMP2
	movlw	b'00011000'		; Initialise high byte of string pointer to point at program bank 3
	movwf	LCDTEMP3
	
	movfw	LCDTEMP2		; Multiply string pointer by 8
	addwf	LCDTEMP2,f		; x2
	btfsc	STATUS,C
	incf	LCDTEMP3
	addwf	LCDTEMP2,f		; x3
	btfsc	STATUS,C
	incf	LCDTEMP3
	addwf	LCDTEMP2,f		; x4
	btfsc	STATUS,C
	incf	LCDTEMP3
	addwf	LCDTEMP2,f		; x5
	btfsc	STATUS,C
	incf	LCDTEMP3
	addwf	LCDTEMP2,f		; x6
	btfsc	STATUS,C
	incf	LCDTEMP3
	addwf	LCDTEMP2,f		; x7
	btfsc	STATUS,C
	incf	LCDTEMP3
	addwf	LCDTEMP2,f		; x8
	btfsc	STATUS,C
	incf	LCDTEMP3
	
	movlw	D'8'			; Count 8 characters
	movwf	LCDTEMP4

LCD_WriteStringFromTEMP2_loop
	call	LCD_WriteStringFromTEMP2_getChar	; Get character from table
	clrf	PCLATH			; Reset bits 3 and 4 of PCLATH (we're in program bank 0 here)
	call	LCD_WriteDR		; Write character to LCD
	
	movlw	1			; Point to next character
	addwf	LCDTEMP2,f	
	btfsc	STATUS,C
	incf	LCDTEMP3
	
	decfsz	LCDTEMP4			; Loop until LCDTEMP4=0
	goto 	LCD_WriteStringFromTEMP2_loop
	return

LCD_WriteStringFromTEMP2_getChar
	movfw	LCDTEMP3
	movwf	PCLATH
	movfw	LCDTEMP2
	movwf	PCL

	
; --------------------------------
